.sphinx-codeautolink-a:link{
    color: inherit;
    text-decoration: none;
}
.sphinx-codeautolink-a:active{
    color: inherit;
    text-decoration: none;
}
.sphinx-codeautolink-a:visited{
    color: inherit;
    text-decoration: none;
}
.sphinx-codeautolink-a:hover{
    color: rgb(0, 139, 139);
    text-decoration: none;
}
